# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""In-memory message queue and result store."""
import logging
from Queue import Queue, Empty
from threading import Lock
from weakref import WeakValueDictionary

import worq.const as const
from worq.core import AbstractTaskQueue

log = logging.getLogger(__name__)

_REFS = WeakValueDictionary()

class TaskQueue(AbstractTaskQueue):
    """Simple in-memory task queue implementation"""

    @classmethod
    def factory(cls, url, name=const.DEFAULT, *args, **kw):
        obj = _REFS.get((url, name))
        if obj is None:
            obj = _REFS[(url, name)] = cls(url, name, *args, **kw)
        return obj

    def __init__(self, *args, **kw):
        super(TaskQueue, self).__init__(*args, **kw)
        self.queue = Queue()
        self.results = WeakValueDictionary()
        self.result_lock = Lock()

    def _init_result(self, result, status, message):
        with self.result_lock:
            if result.id in self.results:
                return False
            self.results[result.id] = result
        result.__status = status
        result.__value = Queue()
        result.__task = message
        result.__args = {}
        result.__lock = Lock()
        result.__for = None
        return True

    def enqueue_task(self, result, message):
        if self._init_result(result, const.ENQUEUED, message):
            self.queue.put(result)
            return True
        return False

    def defer_task(self, result, message, args):
        if self._init_result(result, const.PENDING, message):
            results = self.results
            # keep references to results to prevent GC
            result.__refs = [results[arg] for arg in args]
            return True
        return False

    def undefer_task(self, task_id):
        result = self.results[task_id]
        self.queue.put(result)

    def get(self, timeout=None):
        try:
            result = self.queue.get(timeout=timeout)
        except Empty:
            return None
        result.__status = const.PROCESSING
        return result.id, result.__task

    def discard_pending(self):
        while True:
            try:
                self.queue.get_nowait()
            except Empty:
                break
        with self.result_lock:
            self.results.clear()

    def reserve_argument(self, argument_id, deferred_id):
        result = self.results.get(argument_id)
        if result is None:
            return (False, None)
        with result.__lock:
            if result.__for is not None:
                return (False, None)
            result.__for = deferred_id
            try:
                message = result.__value.get_nowait()
            except Empty:
                message = None
            return (True, message)

    def set_argument(self, task_id, argument_id, message):
        result = self.results[task_id]
        with result.__lock:
            result.__args[argument_id] = message
            return len(result.__args) == len(result.__refs)

    def get_arguments(self, task_id):
        try:
            return self.results[task_id].__args
        except KeyError:
            return {}

    def set_task_timeout(self, task_id, timeout):
        pass

    def get_status(self, task_id):
        result = self.results.get(task_id)
        return None if result is None else result.__status

    def set_result(self, task_id, message, timeout):
        result = self.results.get(task_id)
        if result is not None:
            with result.__lock:
                result.__value.put(message)
                return result.__for

    def pop_result(self, task_id, timeout):
        result = self.results.get(task_id)
        if result is None:
            return const.TASK_EXPIRED
#        with result.__lock:
#            if result.__for is not None:
#                raise NotImplementedError
#                #return const.RESERVED
#            result.__for = task_id
        try:
            if timeout == 0:
                value = result.__value.get_nowait()
            else:
                value = result.__value.get(timeout=timeout)
        except Empty:
            value = None
        else:
            self.results.pop(task_id, None)
        return value

    def discard_result(self, task_id, task_expired_token):
        result = self.results.pop(task_id)
        if result is not None:
            result.__value.put(task_expired_token)



# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import worq.const as const
from worq import get_broker, get_queue, Task, TaskFailure, TaskSpace
from worq.tests.test_examples import example
from worq.tests.util import (assert_raises, eq_, eventually,
    thread_worker, TimeoutLock, WAIT)

@example
def simple(url):
    """A simple example demonstrating WorQ mechanics"""
    state = []

    def func(arg):
        state.append(arg)

    broker = get_broker(url)
    broker.expose(func)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        q.func('arg')

        eventually((lambda:state), ['arg'])


@example
def wait_for_result(url):
    """Efficiently wait for (block on) a task result.

    Use this feature wisely. Waiting for a result in a WorQ task
    could deadlock the queue.
    """

    def func(arg):
        return arg

    broker = get_broker(url)
    broker.expose(func)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        res = q.func('arg')

        completed = res.wait(WAIT)

        assert completed, repr(res)
        eq_(res.value, 'arg')
        eq_(repr(res), "<Deferred func [default:%s] success>" % res.id)


@example
def ignore_result(url):
    """Tell the queue to ignore the task result when the result is not
    important. This is done by creating a ``Task`` object with custom
    options for more efficient queue operation.
    """
    state = []

    def func(arg):
        state.append(arg)

    broker = get_broker(url)
    broker.expose(func)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        f = Task(q.func, ignore_result=True)
        res = f(3)

        eq_(res, None) # verify that we did not get a deferred result
        eventually((lambda:state), [3])


@example
def result_status(url):
    """Deferred results can be queried for task status.

    A lock is used to control state interactions between the producer
    and the worker for illustration purposes only. This type of
    lock-step interaction is not normally needed or even desired.
    """
    lock = TimeoutLock(locked=True)

    def func(arg):
        lock.acquire()
        return arg

    broker = get_broker(url)
    broker.expose(func)
    with thread_worker(broker, lock):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        res = q.func('arg')

        eventually((lambda:res.status), const.ENQUEUED)
        eq_(repr(res), "<Deferred func [default:%s] enqueued>" % res.id)

        lock.release()
        eventually((lambda:res.status), const.PROCESSING)
        eq_(repr(res), "<Deferred func [default:%s] processing>" % res.id)

        lock.release()
        assert res.wait(WAIT), repr(res)
        eq_(repr(res), "<Deferred func [default:%s] success>" % res.id)

        eq_(res.value, 'arg')


@example
def no_such_task(url):

    broker = get_broker(url)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        res = q.func('arg')
        assert res.wait(WAIT), repr(res)

        eq_(repr(res), '<Deferred func [default:%s] failed>' % res.id)
        with assert_raises(TaskFailure,
                'func [default:%s] no such task' % res.id):
            res.value


@example
def task_error(url):

    def func(arg):
        raise Exception('fail!')

    broker = get_broker(url)
    broker.expose(func)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        res = q.func('arg')
        assert res.wait(WAIT), repr(res)

        eq_(repr(res), '<Deferred func [default:%s] failed>' % res.id)
        with assert_raises(TaskFailure,
                'func [default:%s] Exception: fail!' % res.id):
            res.value


@example
def task_with_deferred_arguments(url):
    """A deferred result may be passed as an argument to another task. Tasks
    receiving deferred arguments will not be invoked until the deferred value
    is available. Notice that the value of the deferred argument, not the
    Deferred object itself, is passed to ``sum`` in this example.
    """

    def func(arg):
        return arg

    broker = get_broker(url)
    broker.expose(func)
    broker.expose(sum)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        res = q.sum([
            q.func(1),
            q.func(2),
            q.func(3),
        ])

        assert res.wait(WAIT), repr(res)
        eq_(res.value, 6)


@example
def more_deferred_arguments(url):
    from operator import add

    def func(arg):
        return arg

    broker = get_broker(url)
    broker.expose(func)
    broker.expose(sum)
    broker.expose(add)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        sum_123 = q.sum([
            q.func(1),
            q.func(2),
            q.func(3),
        ])

        sum_1234 = q.add(sum_123, q.func(4))

        assert sum_1234.wait(WAIT), repr(res)
        eq_(sum_1234.value, 10)


@example
def dependency_graph(url):
    """Dependency graph
                         |
            _____________|_____________
           /             |             \
          / \           / \           / \
         /   \         /   \         /   \
      left   right  left   right  left   right
         \   /         \   /         \   /
          \ /           \ /           \ /
         catch         catch         catch
            \            |            /
             \___________|___________/
                         |
                      combine
    """
    ts = TaskSpace()

    @ts.task
    def left(num):
        return ('left', num)

    @ts.task
    def right(num):
        return ('right', num)

    @ts.task
    def catch(left, right, num):
        return [num, left, right]

    @ts.task
    def combine(items):
        return {i[0]: i[1:] for i in items}

    broker = get_broker(url)
    broker.expose(ts)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        catches = []
        for num in [1, 2, 3]:
            left = q.left(num)
            right = q.right(num)

            catch = q.catch(left, right, num)

            catches.append(catch)

        res = q.combine(catches)
        assert res.wait(WAIT), repr(res)

        eq_(res.value, {
            1: [('left', 1), ('right', 1)],
            2: [('left', 2), ('right', 2)],
            3: [('left', 3), ('right', 3)],
        })


@example
def task_with_failed_deferred_arguments(url):
    """TaskFailure can be passed to the final task.

    By default, a task fails if any of its deferred arguments fail. However,
    creating a ``Task`` with ``on_error=Task.PASS`` will cause a ``TaskFailure``
    to be passed as the result of any task that fails.
    """

    def func(arg):
        if arg == 0:
            raise Exception('zero fail!')
        return arg

    broker = get_broker(url)
    broker.expose(func)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        items = [
            q.func(1),
            q.func(0),
            q.func(2),
        ]

        task = Task(q.func, on_error=Task.PASS)
        res = task(items)
        res.wait(timeout=WAIT)

        fail = TaskFailure(
            'func', 'default', items[1].id, 'Exception: zero fail!')
        eq_(res.value, [1, fail, 2])


@example
def named_queue(url):
    """Named queues facilitate discrete queues on a single backend."""

    foo_state = []
    def foo_func(arg):
        foo_state.append(arg)
    foo_broker = get_broker(url, 'foo')
    foo_broker.expose(foo_func)

    bar_state = []
    def bar_func(arg):
        bar_state.append(arg)
    bar_broker = get_broker(url, 'bar')
    bar_broker.expose(bar_func)

    with thread_worker(foo_broker), thread_worker(bar_broker):

        # -- task-invoking code, usually another process --
        f = get_queue(url, 'foo')
        f.foo_func(1)

        b = get_queue(url, 'bar')
        b.bar_func(2)

        eventually((lambda:(foo_state, bar_state)), ([1], [2]))


@example
def task_namespaces(url):
    """Task namepsaces are used to arrange tasks similar to the Python
    package/module hierarchy.
    """

    state = set()
    __name__ = 'module.path'

    ts = TaskSpace(__name__)

    @ts.task
    def foo():
        state.add('foo')

    @ts.task
    def bar(arg):
        state.add(arg)

    broker = get_broker(url)
    broker.expose(ts)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        q.module.path.foo()
        q.module.path.bar(1)

        eventually((lambda:state), {'foo', 1})


@example
def more_namespaces(url):
    state = set()

    foo = TaskSpace('foo')
    bar = TaskSpace('foo.bar')
    baz = TaskSpace('foo.bar.baz')

    @foo.task
    def join(arg):
        state.add('foo-join %s' % arg)

    @bar.task
    def kick(arg):
        state.add('bar-kick %s' % arg)

    @baz.task
    def join(arg):
        state.add('baz-join %s' % arg)

    @baz.task
    def kick(arg):
        state.add('baz-kick %s' % arg)

    broker = get_broker(url)
    broker.expose(foo)
    broker.expose(bar)
    broker.expose(baz)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        q.foo.join(1)
        q.foo.bar.kick(2)
        q.foo.bar.baz.join(3)
        q.foo.bar.baz.kick(4)

        eventually((lambda:state), {
            'foo-join 1',
            'bar-kick 2',
            'baz-join 3',
            'baz-kick 4',
        })


@example
def expose_method(url):
    """Object methods can be exposed too, not just functions."""

    class Database(object):
        """stateful storage"""
        value = None
        def update(self, value):
            self.value = value

    class TaskObj(object):
        """object with task definitions"""
        def __init__(self, db):
            self.db = db
        def update_value(self, value):
            self.db.update(value)

    db = Database()
    obj = TaskObj(db)
    broker = get_broker(url)
    broker.expose(obj.update_value)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)
        q.update_value(2)
        eventually((lambda:db.value), 2)



import logging
import logging.config
import os
import signal
import sys
import time
from multiprocessing import Process, Pipe
from threading import Thread

logging.config.dictConfig({
    'formatters': {
        'brief': {
            'format': '%(levelname)-7s %(name)s - %(message)s',
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'brief',
        },
    },
    'root': {
        'level': 'DEBUG',
        'handlers': ['console'],
    },
    #'disable_existing_loggers': False,
    'version': 1,
})

log = logging.getLogger()

def master():
    results = []
    mcn, scn = Pipe()
    mon = Thread(target=monitor, args=(mcn, results))
    mon.daemon = True
    mon.start()

    proc = Process(target=slave, args=(scn,))
    proc.start()

    mcn.send('1')
    while not results:
        time.sleep(0.01)
    pid = results.pop()
    log.info('pid: %s', pid)

    mcn.send('die')
    try:
        for n in range(7):
            if mcn.poll(1):
                break
            if not proc.is_alive():
                raise Exception('stop')
    except Exception:
        log.error('slave died', exc_info=True)

    #os.kill(pid, signal.SIGKILL)
    #mcn.poll()
    #proc.terminate()

def monitor(cn, results):
    try:
        while True:
            results.append(cn.recv())
            log.info('monitor got %s', results[0])
    except:
        log.error('montor died', exc_info=True)


def slave(cn):
    try:
        while True:
            item = cn.recv()
            log.info('slave got %s' % item)
            if item == 'die':
                sys.exit()
            else:
                cn.send(os.getpid())
    except:
        log.error('slave died', exc_info=True)

if __name__ == '__main__':
    master()



# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Multi-process worker pool

Processes in the ``worq.pool.process`` stack:

* Queue - enqueues tasks to be executed
* Broker - task queue and results backend (redis)
* WorkerPool - worker pool manager process
* Worker - worker process, which does the real work
"""

import errno
import logging
import os
import sys
import time
import subprocess
from functools import partial
from multiprocessing import Pipe, cpu_count, current_process
from multiprocessing.process import AuthenticationString
from multiprocessing.reduction import reduce_connection, rebuild_connection
from cPickle import dump, load, HIGHEST_PROTOCOL, PicklingError
from worq import get_broker
from worq.core import DAY
from Queue import Empty, Queue as ThreadQueue
from threading import Thread

log = logging.getLogger(__name__)

STOP = 'STOP'
PYTHON_EXE = sys.executable
WORKER_POLL_INTERVAL = 30

class Error(Exception): pass


class WorkerPool(object):
    """Multi-process worker pool

    :param broker: Broker object.
    :param init_func: Worker initializer. This is called to initialize
        each worker on startup. It can be used to setup logging or do
        any other global initialization. The first argument will be a
        broker url, and the remaining will be ``*init_args,
        **init_kwargs``. It must return a broker instance, and it must
        be pickleable.
    :param init_args: Additional arguments to pass to init_func.
    :param init_kwargs: Additional keyword arguments to pass to init_func.
    :param workers: Number of workers to maintain in the pool. The default
        value is the number returned by ``multiprocessing.cpu_count``.
    :param max_worker_tasks: Maximum number of tasks to execute on each worker
        before retiring the worker and spawning a new one in its place.
    """

    def __init__(self, broker,
            init_func,
            init_args=(),
            init_kwargs=None,
            workers=None,
            max_worker_tasks=None,
        ):
        self.broker = broker
        self.init_func = init_func
        self.init_args = (broker.url,) + init_args
        self.init_kwargs = {} if init_kwargs is None else init_kwargs
        if workers is None:
            try:
                workers = cpu_count()
            except NotImplementedError:
                workers = 1
        self.workers = workers
        self.max_worker_tasks = max_worker_tasks
        self._workers = []
        self._worker_queue = ThreadQueue()
        self._running = False

    def start(self, timeout=10, handle_sigterm=True):
        """Start the worker pool

        :param timeout: Number of seconds to block while waiting for a
            new task to arrive on the queue. This timeout value affects
            pool stop time: a larger value means shutdown may take
            longer because it may need to wait longer for the consumer
            thread to complete.
        :param handle_sigterm: If true (the default) setup a signal
            handler and block until the process is signalled. This
            should only be called in the main thread in that case. If
            false, start workers and a pool manager thread and return.
        """
        if self._running:
            raise Error('cannot start already running WorkerPool')
        self._running = True
        log.info('starting worker pool...')

        for n in range(self.workers):
            worker = WorkerProxy(
                self.broker,
                self.init_func,
                self.init_args,
                self.init_kwargs,
                self.max_worker_tasks
            )
            self._workers.append(worker)
            self._worker_queue.put(worker)

        args = (timeout,)
        self._consumer_thread = Thread(target=self._consume_tasks, args=args)
        self._consumer_thread.start()
        log.info('WorkerPool-%s started', os.getpid())

        if handle_sigterm:
            log.info('Press CTRL+C or send SIGTERM to stop')
            setup_exit_handler()
            try:
                # Sleep indefinitely waiting for a signal
                # to initiate pool shutdown.
                while True:
                    time.sleep(DAY)
            finally:
                self.stop()

    def _consume_tasks(self, timeout):
        get_task = self.broker.next_task
        get_worker = self._worker_queue.get
        put_worker = self._worker_queue.put
        try:
            while True:
                worker = get_worker()
                if worker == STOP:
                    break
                task = get_task(timeout)
                if task is None:
                    put_worker(worker)
                    continue
                worker.execute(task, put_worker)
        except Exception:
            log.error('task consumer crashed', exc_info=True)
        log.debug('consumer stopped')

    def stop(self):
        """Shutdown the pool

        This is probably only useful when the pool was started with
        `handle_sigterm=False`.
        """
        if not self._running:
            return False
        self._running = False

        log.info('shutting down...')
        while True:
            try:
                item = self._worker_queue.get_nowait()
            except Empty:
                break
        self._worker_queue.put(STOP)
        self._consumer_thread.join()
        self._worker_queue = ThreadQueue()
        for worker in self._workers:
            worker.terminate()
        log.info('worker pool stopped')
        return True


class WorkerProxy(object):

    def __init__(self, *args):
        self.pid = 'not-started'
        self.queue = ThreadQueue()
        self.thread = Thread(target=self._proxy_loop, args=args)
        self.thread.start()

    def execute(self, task, return_to_pool):
        self.queue.put((task, return_to_pool))

    def terminate(self):
        self.queue.put((STOP, None))
        self.queue = None
        self.thread.join()

    def _proxy_loop(self, broker, *args):
        is_debug = partial(log.isEnabledFor, logging.DEBUG)
        pid = os.getpid()
        proc = None
        queue = self.queue

        def stop():
            try:
                if proc is not None:
                    child.send(STOP)
                    child.close()
                    proc.join()
            except Exception:
                log.error('%s failed to stop cleanly',
                    str(self), exc_info=True)
                raise
            else:
                log.debug('terminated %s', str(self))
            finally:
                self.pid = '%s-terminated' % self.pid

        while True:
            if proc is None or not proc.is_alive():
                # start new worker process
                child, parent = Pipe()
                cx = _reduce_connection(parent) # HACK reduce for pickle
                proc = run_in_subprocess(worker_process, pid, cx, *args)
                self.pid = proc.pid

            task, return_to_pool = queue.get()
            if task == STOP:
                stop()
                break

            try:
                child.send(task)
                while not child.poll(task.heartrate):
                    if not proc.is_alive():
                        broker.task_failed(task)
                        raise Error('unknown cause of death')
                    broker.heartbeat(task)
                (result, status) = child.recv()
                broker.set_result(task, result)
            except Exception:
                log.error('%s died unexpectedly', str(self), exc_info=True)
                child.close()
                proc.stdin.close()
                proc = None
            else:
                if is_debug():
                    log.debug('%s completed task', str(self))
                if status == STOP:
                    child.close()
                    proc.stdin.close()
                    proc = None
            finally:
                return_to_pool(self)

    def __str__(self):
        return 'WorkerProxy-%s' % self.pid


def worker_process(parent_pid, reduced_cn,
        init, init_args, init_kw, max_worker_tasks):
    broker = init(*init_args, **init_kw)
    log.info('Worker-%s started', os.getpid())
    task_count = 1
    parent = reduced_cn[0](*reduced_cn[1]) # HACK un-reduce connection

    while True:
        while not parent.poll(WORKER_POLL_INTERVAL):
            if os.getppid() != parent_pid:
                log.error('abort: parent process changed')
                return

        try:
            task = parent.recv()
        except EOFError:
            log.error('abort: parent fd closed')
            break
        if task == STOP:
            break

        result = broker.invoke(task, return_result=True)

        if max_worker_tasks is None:
            parent.send((result, None)) # send result
        elif task_count < max_worker_tasks:
            task_count += 1
            parent.send((result, None)) # send result
        else:
            parent.send((result, STOP)) # send result, worker stopping
            break

    log.info('Worker-%s stopped', os.getpid())


def run_in_subprocess(_func, *args, **kw):
    """Call function with arguments in subprocess

    All arguments to this function must be able to be pickled.

    Use ``subprocess.Popen`` rather than ``multiprocessing.Process``
    because we use threads, which do not play nicely with ``fork``. This
    was originally written with ``multiprocessing.Process``, which caused
    in intermittent deadlocks. See http://bugs.python.org/issue6721

    :returns: A ``PopenProcess`` object.
    """
    prog = 'from worq.pool.process import main; main()'
    # close_fds=True prevents intermittent deadlock in Popen
    # See http://bugs.python.org/issue2320
    proc = subprocess.Popen([PYTHON_EXE, '-c', prog],
                            stdin=subprocess.PIPE, close_fds=True,
                            preexec_fn=disable_signal_propagation)
    assert proc.stdout is None
    assert proc.stderr is None
    try:
        dump((_func, args, kw), proc.stdin, HIGHEST_PROTOCOL)
        proc.stdin.flush()
    except IOError as e:
        # copied from subprocess.Popen.communicate
        if e.errno != errno.EPIPE and e.errno != errno.EINVAL:
            proc.terminate()
            raise
    except PicklingError:
        proc.terminate()
        raise
    return PopenProcess(proc)


def main():
    func, args, kw = load(sys.stdin)
    try:
        func(*args, **kw)
    except Exception:
        log.critical('subprocess crashed', exc_info=True)
        raise


def _reduce_connection(conn):
    """Reduce a connection object so it can be pickled.

    WARNING this puts the current process' authentication key in the data
    to be pickled. Connections pickled with this function should not be
    sent over an untrusted network.

    HACK work around ``multiprocessing`` connection authentication because
    we are using ``subprocess.Popen`` instead of ``multiprocessing.Process``
    to spawn new child processes.

    This will not be necessary when ``multiprocessing.Connection`` objects
    can be pickled. See http://bugs.python.org/issue4892
    """
    obj = reduce_connection(conn)
    assert obj[0] is rebuild_connection, obj
    assert len(obj) == 2, obj
    args = (bytes(current_process().authkey),) + obj[1]
    return (_rebuild_connection, args)

def _rebuild_connection(authkey, *args):
    current_process().authkey = AuthenticationString(authkey)
    return rebuild_connection(*args)


class PopenProcess(object):
    """Make a ``subprocess.Popen`` object more like ``multiprocessing.Process``
    """

    def __init__(self, proc):
        self._proc = proc

    def is_alive(self):
        return self._proc.poll() is None

    def join(self, timeout=None):
        if timeout is None:
            self._proc.communicate()
            return
        end = time.time() + timeout
        while time.time() < end:
            if not self.is_alive():
                return
            time.sleep(0.01)

    def __getattr__(self, name):
        return getattr(self._proc, name)


def disable_signal_propagation():
    # http://stackoverflow.com/a/5446983/10840
    os.setpgrp()


def setup_exit_handler():
    # http://danielkaes.wordpress.com/2009/06/04/how-to-catch-kill-events-with-python/
    def on_exit(sig, func=None):
        sys.exit()
    if os.name == "nt":
        try:
            import win32api
            win32api.SetConsoleCtrlHandler(on_exit, True)
        except ImportError:
            version = ".".join(map(str, sys.version_info[:2]))
            raise Exception("pywin32 not installed for Python " + version)
    else:
        import signal
        signal.signal(signal.SIGINT, on_exit)
        signal.signal(signal.SIGTERM, on_exit)



# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import time
from uuid import uuid4

from worq.const import HOUR, DEFAULT

log = logging.getLogger(__name__)


class Queue(object):
    """Queue for invoking remote tasks

    New ``Queue`` instances are generated through attribute access.
    For example::

        >>> q = Queue(broker)
        >>> q.foo
        <Queue foo [default]>
        >>> q.foo.bar
        <Queue foo.bar [default]>

    A ``Queue`` instance can be called like a function, which invokes a remote
    task identified by the target of the ``Queue`` instance. Example::

        # Enqueue task 'func' in namespace 'foo' to be invoked
        # by a worker listening on the 'default' queue.
        >>> q = Queue(broker)
        >>> q.foo.func(1, key=None)

    The arrangement of queue tasks in ``TaskSpaces`` is similar to Python's
    package/module/function hierarchy.

    NOTE two ``Queue`` objects are considered equal if they refer to the same
    broker (their targets may be different).
    """

    def __init__(self, broker, target=''):
        self.__broker = broker
        self.__target = target

    def __getattr__(self, target):
        if self.__target:
            target = '%s.%s' % (self.__target, target)
        return Queue(self.__broker, target)

    def __call__(self, *args, **kw):
        """Invoke the task identified by this ``Queue``"""
        return Task(self)(*args, **kw)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.__broker == other.__broker

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '<Queue %s [%s]>' % (self.__target, self.__broker.name)

    def __str__(self):
        return self.__target


def option_descriptors(cls):
    def make_getter(name, default):
        return lambda self: self.options.get(name, default)
    for name, default in cls.OPTIONS.iteritems():
        fget = make_getter(name, default)
        setattr(cls, name, property(fget))
    return cls

@option_descriptors
class Task(object):
    """Remote task handle

    This class can be used to construct a task with custom options.
    A task is invoked by calling the task object.

    :param queue: The ``Queue`` object identifying the task to be executed.
    :param id: A unique identifier string for this task, or a function
        that returns a unique identifier string when called with the
        task's arguments. If not specified, a global unique identifier
        is generated for each call. Only one task with a given id may
        exist in the queue at any given time. Note that a task with
        ``ignore_result=True`` will be removed from the queue before it
        is invoked.
    :param on_error: What should happen when a deferred argument's task
        fails. The ``TaskFailure`` exception will be passed as an
        argument if this value is ``Task.PASS``, otherwise this will
        fail before it is invoked (the default action).
    :param ignore_result: Create a fire-and-forget task if true. Task
        invocation will return ``None`` rather than a ``Deferred`` object.
    :param result_timeout: Number of seconds to retain the result after
        the task has completed. The default is one hour. This is ignored
        by some ``TaskQueue`` implementations.
    :param heartrate: Number of seconds between task heartbeats, which
        are maintained by some ``WorkerPool`` implementations to prevent
        result timeout while the task is running. The default is 30
        seconds.
    """

    PASS = 'pass'
    FAIL = 'fail'

    OPTIONS = {
        'on_error': FAIL,
        'ignore_result': False,
        'result_timeout': HOUR,
        'heartrate': 30,
    }

    def __init__(self, queue,
                id=None,
                on_error=FAIL,
                ignore_result=False,
                result_timeout=HOUR,
                heartrate=30,
            ):
        self.queue = queue
        self.name = queue._Queue__target
        self.options = options = {}

        if id is not None:
            options['id'] = id

        if on_error not in [Task.PASS, Task.FAIL]:
            raise ValueError('invalid on_error: %r' % (on_error,))
        if on_error == Task.PASS:
            options['on_error'] = on_error

        if ignore_result:
            if result_timeout != HOUR:
                raise ValueError(
                    'ignore_result is incompatible with result_timeout')
            options['ignore_result'] = bool(ignore_result)

        if not isinstance(result_timeout, (int, long, float)):
            raise ValueError('invalid result_timeout: %r' % (result_timeout,))
        if result_timeout != HOUR:
            options['result_timeout'] = result_timeout

        if not isinstance(heartrate, (int, long, float)):
            raise ValueError('invalid heartrate: %r' % (heartrate,))
        if heartrate != 30:
            options['heartrate'] = heartrate

    @property
    def broker(self):
        return self.queue._Queue__broker

    def __call__(self, *args, **kw):
        task = FunctionTask(self.name, args, kw, self.options)
        return self.queue._Queue__broker.enqueue(task)

    def with_options(self, options):
        """Clone this task with a new set of options"""
        return Task(self.queue, **options)

    def __repr__(self):
        return '<Task %s [%s]>' % (self.name, self.queue._Queue__queue)


@option_descriptors
class FunctionTask(object):

    OPTIONS = Task.OPTIONS

    def __init__(self, name, args, kw, options):
        if 'id' in options:
            options = dict(options)
            ident = options.pop('id')
            if isinstance(ident, basestring):
                self.id = ident
            else:
                self.id = ident(*args, **kw)
        else:
            self.id = uuid4().hex
        self.name = name
        self.args = args
        self.kw = kw
        self.options = options

    @property
    def on_error_pass(self):
        return self.options.get('on_error') == Task.PASS

    def invoke(self, broker, return_result=False):
        queue = broker.name
        log.debug('invoke %s [%s:%s] %s',
            self.name, queue, self.id, self.options)
        try:
            try:
                task = broker.tasks[self.name]
            except KeyError:
                result = TaskFailure(self.name, queue, self.id, 'no such task')
                log.error(result)
            else:
                result = task(*self.args, **self.kw)
        except Exception, err:
            log.error('task failed: %s [%s:%s]',
                self.name, queue, self.id, exc_info=True)
            result = TaskFailure(self.name, queue, self.id,
                '%s: %s' % (type(err).__name__, err))
        except BaseException, err:
            log.error('worker died in task: %s [%s:%s]',
                self.name, queue, self.id, exc_info=True)
            result = TaskFailure(self.name, queue, self.id,
                '%s: %s' % (type(err).__name__, err))
            raise
        finally:
            if return_result:
                return result
            # hope this doesn't run if an exception has been raised
            broker.set_result(self, result)


class Deferred(object):
    """Deferred result object

    Not thread-safe.
    """

    def __init__(self, broker, task):
        self.broker = broker
        self.task = task
        self._status = None

    @property
    def id(self):
        return self.task.id

    @property
    def name(self):
        return self.task.name

    @property
    def value(self):
        """Get the value returned by the task (if completed)

        :returns: The value returned by the task if it completed successfully.
        :raises: AttributeError if the task has not yet completed. TaskFailure
            if the task failed for any reason.
        """
        self.wait(0)
        try:
            value = self._value
        except AttributeError:
            raise AttributeError('Deferred value not available')
        if isinstance(value, TaskFailure):
            raise value
        return value

    @property
    def status(self):
        """Get task status"""
        if self:
            if isinstance(self._value, TaskFailure):
                self._status = 'failed'
            else:
                self._status = 'success'
        else:
            self._status = self.broker.status(self)
        return self._status

    def wait(self, timeout):
        """Wait for the task result.

        Use this method wisely. In general a task should never wait on the
        result of another task because it may cause deadlock.

        :param timeout: Number of seconds to wait. A value of ``None`` will wait
            indefinitely, but this is dangerous since the worker may go away
            without notice (due to loss of power, etc.) causing this method
            to deadlock.
        :returns: True if the result is available, otherwise False.
        """
        if not hasattr(self, '_value'):
            try:
                value = self.broker.pop_result(self, timeout=timeout)
            except KeyError:
                return False
            except TaskExpired, err:
                value = err
            self._value = value
        return hasattr(self, '_value')

    def __nonzero__(self):
        """Return True if the result has arrived, otherwise False."""
        return self.wait(0)

    def __repr__(self):
        status = self.status
        if status is None:
            status = 'incomplete'
        args = (self.name, self.broker.name, self.id, status)
        return '<Deferred %s [%s:%s] %s>' % args


class TaskSpace(object):
    """Task namespace container"""

    def __init__(self, name=''):
        self.name = name
        self.tasks = {}

    def task(self, callable, name=None):
        """Add a task to the namespace

        This can be used as a decorator::

            ts = TaskSpace(__name__)

            @ts.task
            def frob(value):
                db.update(value)

        :param callable: A callable object, usually a function or method.
        :param name: Task name. Defaults to `callable.__name__`.
        """
        if name is None:
            name = callable.__name__
        if self.name:
            name = '%s.%s' % (self.name, name)
        if name in self.tasks:
            raise ValueError('task %r conflicts with existing task' % name)
        self.tasks[name] = callable
        return callable


class TaskFailure(Exception):
    """Task failure exception class

    Initialize with the following positional arguments:
        1. Task name
        2. Queue name
        3. Task id
        4. Error text
    """

    @property
    def task_name(self):
        return self.args[0]

    @property
    def queue(self):
        return self.args[1]

    @property
    def task_id(self):
        return self.args[2]

    @property
    def error(self):
        return self.args[3]

    def __str__(self):
        return '%s [%s:%s] %s' % self.args

    def __repr__(self):
        return '<%s %s>' % (type(self).__name__, self)

    def __eq__(self, other):
        return isinstance(other, TaskFailure) and repr(self) == repr(other)

    def __ne__(self, other):
        return not self.__eq__(other)


class TaskExpired(TaskFailure): pass



# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from collections import defaultdict
from cPickle import (Pickler, PicklingError, Unpickler, UnpicklingError, loads,
    HIGHEST_PROTOCOL)
from cStringIO import StringIO
from uuid import uuid4
from weakref import ref as weakref

from worq.const import DEFAULT, HOUR, MINUTE, DAY, STATUS_VALUES, TASK_EXPIRED
from worq.task import (Queue, TaskSpace, FunctionTask, Deferred,
    TaskFailure, TaskExpired)

__all__ = ['Broker', 'AbstractTaskQueue']

log = logging.getLogger(__name__)

class Broker(object):
    """A Broker controlls all interaction with the queue backend"""

    def __init__(self, message_queue):
        self.messages = message_queue
        self.tasks = {}
        self.name = message_queue.name

    @property
    def url(self):
        return self.messages.url

    def expose(self, obj, replace=False):
        """Expose a TaskSpace or task callable.

        :param obj: A TaskSpace or task callable.
        :param replace: Replace existing task if True. Otherwise (by default),
            raise ValueError if this would replace an existing task.
        """
        if isinstance(obj, TaskSpace):
            space = obj
        else:
            space = TaskSpace()
            space.task(obj)
        for name, func in space.tasks.iteritems():
            if name in self.tasks and not replace:
                raise ValueError('task %r conflicts with existing task' % name)
            self.tasks[name] = func

    def discard_pending_tasks(self):
        """Discard pending tasks from queue"""
        self.messages.discard_pending()

    def queue(self, target=''):
        """Get a Queue from the broker"""
        return Queue(self, target)

    def enqueue(self, task):
        message, args = self.serialize(task, deferred=True)
        result = Deferred(self, task)
        store = self.messages
        if args:
            if not store.defer_task(result, message, args):
                raise TaskFailure(task.name, self.name, task.id,
                    'cannot enqueue task with duplicate id')
            log.debug('defer %s [%s:%s]', task.name, self.name, task.id)
            for arg_id in args:
                ok, msg = store.reserve_argument(arg_id, task.id)
                if not ok:
                    raise TaskFailure(task.name, self.name, task.id,
                        'task [%s:%s] result is not available'
                        % (self.name, arg_id))
                if msg is not None:
                    allset = store.set_argument(task.id, arg_id, msg)
                    if allset:
                        log.debug('undefer %s [%s:%s]',
                            task.name, self.name, task.id)
                        self.messages.undefer_task(task.id)
        else:
            if not store.enqueue_task(result, message):
                raise TaskFailure(task.name, self.name, task.id,
                    'cannot enqueue task with duplicate id')
            log.debug('enqueue %s [%s:%s]', task.name, self.name, task.id)
        return None if task.ignore_result else result

    def status(self, result):
        """Get the status of a deferred result"""
        message = self.messages.get_status(result.id)
        if message is None:
            return message
        if message in STATUS_VALUES:
            return message
        return self.deserialize(message)

    def next_task(self, timeout=None):
        """Get the next task from the queue.

        :param timeout: See ``AbstractTaskQueue.get``.
        :returns: A task object. ``None`` on timeout expiration or if the task
            could not be deserialized.
        """
        message = self.messages.get(timeout=timeout)
        if message is None:
            return message
        task_id, message = message
        try:
            task = self.deserialize(message, task_id)
        except Exception:
            log.error('cannot deserialize task [%s:%s]',
                self.name, task_id, exc_info=True)
            return None
        #log.debug('next %s [%s:%s]', task.name, self.name, task_id)
        return task

    def invoke(self, task, **kw):
        """Invoke the given task (normally only called by a worker)"""
        return task.invoke(self, **kw)

    def heartbeat(self, task):
        """Extend task result timeout"""
        timeout = task.heartrate * 2 + 5
        self.messages.set_task_timeout(task.id, timeout)
        taskset_id = task.options.get('taskset_id')
        if taskset_id is not None:
            self.messages.set_task_timeout(taskset_id, timeout)

    def serialize(self, obj, deferred=False):
        """Serialize an object

        :param obj: The object to serialize.
        :param deferred: When this is true Deferred objects are
            serialized and their values are loaded on deserialization.
            When this is false Deferred objects are not serializable.
        """
        if deferred:
            args = set()
            def persistent_id(obj):
                if isinstance(obj, Deferred):
                    args.add(obj.id)
                    return obj.id
                return None
        else:
            args = None
            def persistent_id(obj):
                if isinstance(obj, Deferred):
                    raise PicklingError('%s cannot be serialized' % obj)
                return None
        data = StringIO()
        pickle = Pickler(data, HIGHEST_PROTOCOL)
        pickle.persistent_id = persistent_id
        pickle.dump(obj)
        msg = data.getvalue()
        return (msg, args) if deferred else msg

    def deserialize(self, message, task_id=None):
        """Deserialize an object

        :param message: A serialized object (string).
        :param deferred: When true load deferreds. When false
            raise an error if the message contains deferreds.
        """
        fail = []
        if task_id is None:
            persistent_load = []
        else:
            args = self.messages.get_arguments(task_id)
            args = {k: loads(v) for k, v in args.items()}
            def persistent_load(arg_id):
                value = args[arg_id]
                if isinstance(value, TaskFailure):
                    fail.append(value)
                return value
        data = StringIO(message)
        pickle = Unpickler(data)
        pickle.persistent_load = persistent_load
        obj = pickle.load()
        if task_id is None and persistent_load:
            raise UnpicklingError('message contained references to '
                'external objects: %s', persistent_load)
        if fail and not obj.on_error_pass:
            # TODO detect errors earlier, fail earlier, cancel enqueued tasks
            self.set_result(obj, fail[0])
            obj = None
        return obj

    def set_result(self, task, result):
        """Persist result object.

        :param task: Task object for which to set the result.
        :param result: Result object.
        """
        timeout = task.result_timeout
        message = self.serialize(result)
        reserve_id = self.messages.set_result(task.id, message, timeout)
        if reserve_id is not None:
            allset = self.messages.set_argument(reserve_id, task.id, message)
            if allset:
                log.debug('undefer [%s:%s]', self.name, reserve_id)
                self.messages.undefer_task(reserve_id)

    def pop_result(self, task, timeout=0):
        """Pop and deserialize a task's result object

        :param task: An object with ``id`` and ``name`` attributes
            representing the task.
        :param timeout: Length of time to wait for the result. The default
            behavior is to return immediately (no wait). Wait indefinitely
            if ``None``.
        :returns: The deserialized result object.
        :raises: KeyError if the result was not available.
        :raises: TaskExpired if the task expired before a result was returned.
            A task normally only expires if the pool loses its ability
            to communicate with the worker performing the task.
        """
        if timeout < 0:
            raise ValueError('negative timeout not supported')
        message = self.messages.pop_result(task.id, timeout)
        if message is None:
            raise KeyError(task.id)
        if message is TASK_EXPIRED:
            result = message
        else:
            result = self.deserialize(message)
        if result is TASK_EXPIRED:
            raise TaskExpired(task.name, self.name, task.id,
                'task expired before a result was returned')
        return result

    def task_failed(self, task):
        """Signal that the given task has failed."""
        self.messages.discard_result(task.id, self.serialize(TASK_EXPIRED))


class AbstractTaskQueue(object):
    """Message queue abstract base class

    Task/result lifecycle

        1. Atomically store non-expiring result placeholder and enqueue task.
        2. Atomically pop task from queue and set timeout on result placeholder.
        3. Task heartbeats extend result expiration as needed.
        4. Task finishes and result value is saved.

    All methods must be thread-safe.

    :param url: URL used to identify the queue.
    :param name: Queue name.
    """

    def __init__(self, url, name=DEFAULT):
        self.url = url
        self.name = name

    def enqueue_task(self, result, message):
        """Enqueue task

        :param result: A ``Deferred`` result for the task.
        :param message: Serialized task message.
        :returns: True if the task was enqueued, otherwise False
            (duplicate task id).
        """
        raise NotImplementedError('abstract method')

    def defer_task(self, result, message, args):
        """Defer a task until its arguments become available

        :param result: A ``Deferred`` result for the task.
        :param message: The serialized task message.
        :param args: A list of task identifiers whose results will be
            included in the arguments to the task.
        """
        raise NotImplementedError('abstract method')

    def undefer_task(self, task_id):
        """Enqueue a deferred task

        All deferred arguments must be available immediately.
        """
        raise NotImplementedError('abstract method')

    def get(self, timeout=None):
        """Atomically get a serialized task message from the queue

        Task processing has started when this method returns, which
        means that the task heartbeat must be maintained if there could
        be someone waiting on the result. The result status is set to
        ``worq.const.PROCESSING`` if a result is being maintained for
        the task.

        :param timeout: Number of seconds to wait before returning ``None`` if
            no task is available in the queue. Wait forever if timeout is
            ``None``.
        :returns: A two-tuple (<task_id>, <serialized task message>) or
            ``None`` if timeout was reached before a task arrived.
        """
        raise NotImplementedError('abstract method')

    def discard_pending(self):
        """Discard pending tasks from queue"""
        raise NotImplementedError('abstract method')

    def reserve_argument(self, argument_id, deferred_id):
        """Reserve the result of a task as an argument of a deferred task

        :param argument_id: Identifier of a task whose result will
            be reserved for another task.
        :param deferred_id: Identifier of a deferred task who will get
            the reserved result as an argument.
        :returns: A two-tuple: (<bool>, <str>). The first item is a flag
            denoting if the argument was reserved, and the second is
            the serialized result if it was available else ``None``.
        """
        raise NotImplementedError('abstract method')

    def set_argument(self, task_id, argument_id, message):
        """Set deferred argument for task

        :param task_id: The identifier of the task to which the argument will
            be passed.
        :param argument_id: The argument identifier.
        :param message: The serialized argument value.
        :returns: True if all arguments have been set for the task.
        """
        raise NotImplementedError('abstract method')

    def get_arguments(self, task_id):
        """Get a dict of deferred arguments

        :param task_id: The identifier of the task to which the arguments will
            be passed.
        :returns: A dict of serialized arguments keyed by argument id.
        """
        raise NotImplementedError('abstract method')

    def set_task_timeout(self, task_id, timeout):
        """Set a timeout on the task result

        Recursively set the timeout on the given task and all deferred
        tasks depending on this task's result.
        """
        raise NotImplementedError('abstract method')

    def get_status(self, task_id):
        """Get the status of a task

        :param task_id: Unique task identifier string.
        :returns: A serialized task status object or ``None``.
        """
        raise NotImplementedError('abstract method')

    def set_result(self, task_id, message, timeout):
        """Persist serialized result message.

        This also sets the result status to ``worq.const.COMPLETED``.

        :param task_id: Unique task identifier string.
        :param message: Serialized result message.
        :param timeout: Number of seconds to persist the result before
            discarding it.
        :returns: A deferred task identifier if the result has been reserved.
            Otherwise ``None``.
        """
        raise NotImplementedError('abstract method')

    def pop_result(self, task_id, timeout):
        """Pop serialized result message from persistent storage.

        :param task_id: Unique task identifier string.
        :param timeout: Number of seconds to wait for the result. Wait
            indefinitely if ``None``. Return immediately if timeout is zero (0).
        :returns: One of the following:

            * The result message.
            * ``worq.const.RESERVED`` if another task depends on the result.
            * ``worq.const.TASK_EXPIRED`` if the task expired before a
              result was available.
            * ``None`` on timeout.

        """
        raise NotImplementedError('abstract method')

    def discard_result(self, task_id, task_expired_token):
        """Discard the result for the given task.

        A call to ``pop_result`` after this is invoked should return a
        task expired response.

        :param task_id: The task identifier.
        :param task_expired_token: A message that can be sent to blocking
            actors to signify that the task has expired.
        """
        raise NotImplementedError('abstract method')



import os

TEST_URLS = ['memory://']

def setup():
    for get_test_url in [get_redis_url]:
        url = get_test_url()
        if url is not None:
            TEST_URLS.append(url)

def get_redis_url():
    redis_url = os.environ.get(
        'WORQ_TEST_REDIS_URL', 'redis://localhost:16379/0') # non-standard port
    if redis_url != 'disabled':
        try:
            from worq.queue.redis import TaskQueue as RedisQueue
        except ImportError:
            pass
        else:
            try:
                queue = RedisQueue(redis_url)
                if queue.ping():
                    return redis_url
            except Exception:
                pass
    return None

def test_redis_should_be_installed():
    if 'WORQ_TEST_REDIS_URL' in os.environ:
        try:
            import redis
        except ImportError:
            assert 0, 'WORQ_TEST_REDIS_URL is set but redis is not installed'



# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from threading import Event, Thread
from uuid import uuid4
from worq.task import TaskSpace

log = logging.getLogger(__name__)


class WorkerPool(object):
    """Multi-thread worker pool

    :param broker: Queue broker instance.
    :param workers: Number of workers in the pool.
    :param thread_factory: A factory function that creates a new thread
        object. This should have the same signature as
        ``threading.Thread`` and should return a thread object.
    """

    def __init__(self, broker, workers=1, thread_factory=Thread):
        self.broker = broker
        self.workers = workers
        self.thread_factory = thread_factory
        self.threads = []
        self.stop_event = Event()
        ts = TaskSpace(__name__)
        ts.task(lambda:None, 'noop')
        broker.expose(ts, replace=True)

    def start(self, timeout=1):
        """Start worker threads.

        :param timeout: The number of seconds to wait for a task before
            checking if the pool has been asked to stop.
        """
        self.stop_event.clear()
        for ident in xrange(self.workers):
            args = (ident, self.broker, timeout, self.stop_event)
            thread = self.thread_factory(target=worker, args=args)
            thread.start()
            self.threads.append(thread)

    def stop(self, use_sentinel=False, join=True):
        """Stop the worker pool

        :param use_sentinel: Enqueue a no-op task for each worker if true.
            This will result in a more responsive shutdown if there are no
            other worker pools consuming tasks from the broker.
        :param join: Join each thread afer sending the stop signal.
        """
        self.stop_event.set()
        if use_sentinel:
            q = self.broker.queue(__name__)
            for thread in self.threads:
                if thread.is_alive():
                    q.noop()
        if join:
            self.join()

    def join(self):
        """Wait for all threads to stop (call ``stop(join=False)`` first)"""
        for thread in self.threads:
            thread.join()


def worker(ident, broker, timeout, stop_event):
    try:
        while not stop_event.is_set():
            task = broker.next_task(timeout=timeout)
            if task is not None:
                broker.invoke(task)
    except:
        log.error('worker %s crashed', ident, exc_info=True)



# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import functools
import time
import logging
import shutil
from contextlib import contextmanager
from tempfile import mkdtemp
from threading import Lock

from worq.pool.thread import WorkerPool
from worq.tests import TEST_URLS

log = logging.getLogger(__name__)

WAIT = 5 #30 # default test timeout (seconds)
DEFAULT_TIMEOUT = 10 # default utility timeout

def with_urls(test=None, exclude=None):
    if test is not None:
        @functools.wraps(test)
        def wrapper():
            for url in TEST_URLS:
                if exclude:
                    if not url.startswith(exclude):
                        yield test, url
                else:
                    yield test, url
        return wrapper
    return functools.partial(with_urls, exclude=exclude)

@contextmanager
def thread_worker(broker, lock=None, timeout=1):
    if lock is not None:
        real_next_task = broker.next_task
        def next_task(*args, **kw):
            log.debug('acquiring lock before getting next task')
            lock.acquire()
            return real_next_task(*args, **kw)
        broker.next_task = next_task
    pool = WorkerPool(broker)
    pool.start(timeout=timeout)
    try:
        yield
    finally:
        pool.stop(use_sentinel=True, join=False)
        try:
            if lock is not None:
                lock.release()
        except Exception:
            log.error('lock release failed', exc_info=True)
        finally:
            pool.join()
            broker.discard_pending_tasks()

class TimeoutLock(object):
    """A lock with acquisition timeout; initially locked by default."""
    def __init__(self, locked=False):
        self.lock = Lock()
        if locked:
            self.lock.acquire()
    def acquire(self, timeout=DEFAULT_TIMEOUT):
        end = time.time() + timeout
        while time.time() < end:
            if self.lock.acquire(False):
                return
        raise Exception('lock timeout')
    def release(self):
        self.lock.release()

def eventually(get_value, expect, timeout=DEFAULT_TIMEOUT, poll_interval=0):
    end = time.time() + timeout
    while time.time() < end:
        actual = get_value()
        if actual == expect:
            return
        time.sleep(poll_interval)
    raise AssertionError('eventually timeout: %r != %r' % (actual, expect))

@contextmanager
def tempdir(*args, **kw):
    """Create a temporary directory to be used in a test

    If (optional keyword argument) 'delete' evaluates to True (the default
    value), the temporary directory and all files in it will be removed on
    context manager exit.
    """
    delete = kw.pop("delete", True)
    path = mkdtemp(*args, **kw)
    try:
        yield path
    finally:
        if delete:
            shutil.rmtree(path)

@contextmanager
def assert_raises(exc_class, msg=None):
    try:
        yield
    except exc_class as exc:
        if isinstance(msg, basestring):
            eq_(str(exc), msg)
        elif msg is not None:
            msg.search(str(exc))
    else:
        raise AssertionError('%s not raised' % exc_class.__name__)

def eq_(value, other):
    assert value == other, '%r != %r' % (value, other)



# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Redis message queue and result store."""
from __future__ import absolute_import
import logging
import redis
import time
import worq.const as const
from urlparse import urlparse
from worq.core import AbstractTaskQueue

log = logging.getLogger(__name__)

QUEUE_PATTERN = 'worq:queue:%s'
TASK_PATTERN = 'worq:task:%s:%s'
ARGS_PATTERN = 'worq:args:%s:%s'
RESERVE_PATTERN = 'worq:reserved:%s:%s'
RESULT_PATTERN = 'worq:result:%s:%s'


class TaskQueue(AbstractTaskQueue):
    """Redis task queue"""

    def __init__(self, url, name=const.DEFAULT, initial_result_timeout=60,
            redis_factory=redis.StrictRedis):
        super(TaskQueue, self).__init__(url, name)
        urlobj = urlparse(url)
        if ':' in urlobj.netloc:
            host, port = urlobj.netloc.rsplit(':', 1)
        else:
            host, port = urlobj.netloc, 6379
        db = int(urlobj.path.lstrip('/'))
        self.redis = redis_factory(host, int(port), db=db)
        self.queue_key = QUEUE_PATTERN % self.name
        self.initial_result_timeout = max(int(initial_result_timeout), 1)

    def ping(self):
        return self.redis.ping()

    def enqueue_task(self, result, message):
        task_key = TASK_PATTERN % (self.name, result.id)
        result_key = RESULT_PATTERN % (self.name, result.id)
        with self.redis.pipeline() as pipe:
            try:
                pipe.watch(task_key)
                if pipe.exists(task_key):
                    return False
                pipe.multi()
                pipe.hmset(task_key, {
                    'task': message,
                    'status': const.ENQUEUED
                })
                pipe.delete(result_key)
                pipe.lpush(self.queue_key, result.id) # left push (head)
                pipe.execute()
                return True
            except redis.WatchError:
                return False

    def defer_task(self, result, message, args):
        task_key = TASK_PATTERN % (self.name, result.id)
        result_key = RESULT_PATTERN % (self.name, result.id)
        with self.redis.pipeline() as pipe:
            try:
                pipe.watch(task_key)
                if pipe.exists(task_key):
                    return False
                pipe.multi()
                pipe.hmset(task_key, {
                    'task': message,
                    'status': const.PENDING,
                    'total_args': len(args),
                    #'args_ready': 0, zero by default
                })
                pipe.delete(result_key)
                pipe.execute()
                return True
            except redis.WatchError:
                return False

    def undefer_task(self, task_id):
        self.redis.lpush(self.queue_key, task_id)

    def get(self, timeout=0):
        queue = self.queue_key
        end = time.time() + timeout
        while True:
            task_id = self.redis.brpoplpush(queue, queue, timeout=timeout)
            if task_id is None:
                return task_id # timeout

            key = TASK_PATTERN % (self.name, task_id)
            with self.redis.pipeline() as pipe:
                pipe.hget(key, 'task')
                pipe.hset(key, 'status', const.PROCESSING)
                pipe.expire(key, self.initial_result_timeout)
                pipe.lrem(queue, 1, task_id) # traverse head to tail
                message, x, x, removed = pipe.execute()
            if removed:
                return (task_id, message)

            if timeout != 0:
                timeout = int(end - time.time())
                if timeout <= 0:
                    return None

    def discard_pending(self):
        with self.redis.pipeline() as pipe:
            pipe.lrange(self.queue_key, 0, -1)
            pipe.delete(self.queue_key)
            tasks = pipe.execute()[0]
        if tasks:
            self.redis.delete(*(TASK_PATTERN % (self.name, t) for t in tasks))

    def reserve_argument(self, argument_id, deferred_id):
        reserve_key = RESERVE_PATTERN % (self.name, argument_id)
        result_key = RESULT_PATTERN % (self.name, argument_id)
        with self.redis.pipeline() as pipe:
            try:
                pipe.watch(reserve_key)
                value = pipe.get(reserve_key)
                if value is not None:
                    return (False, None)
                pipe.multi()
                pipe.set(reserve_key, deferred_id)
                pipe.lpop(result_key)
                result = pipe.execute()[-1]
                return (True, result)
            except redis.WatchError:
                return (False, None)

    def set_argument(self, task_id, argument_id, message):
        task_key = TASK_PATTERN % (self.name, task_id)
        args_key = ARGS_PATTERN % (self.name, task_id)
        with self.redis.pipeline() as pipe:
            pipe.hincrby(task_key, 'args_ready', 1)
            pipe.hget(task_key, 'total_args')
            pipe.hset(args_key, argument_id, message)
            ready, total, x = pipe.execute()
        total = int(total)
        assert isinstance(ready, (int, long)), repr(ready)
        assert ready > 0, ready
        assert total > 0, total
        return ready == total

    def get_arguments(self, task_id):
        args_key = ARGS_PATTERN % (self.name, task_id)
        with self.redis.pipeline() as pipe:
            pipe.hgetall(args_key)
            pipe.delete(args_key)
            return pipe.execute()[0] or {}

    def set_task_timeout(self, task_id, timeout):
        def set_timeout(task_id, args=True, timeout=max(int(timeout), 1)):
            task_key = TASK_PATTERN % (self.name, task_id)
            reserve_key = RESERVE_PATTERN % (self.name, task_id)
            with self.redis.pipeline() as pipe:
                while True:
                    try:
                        pipe.watch(reserve_key)
                        deferred_id = pipe.get(reserve_key)
                        pipe.multi()
                        pipe.expire(task_key, timeout)
                        if args:
                            args_key = ARGS_PATTERN % (self.name, task_id)
                            pipe.expire(args_key, timeout)
                        pipe.execute()
                        return deferred_id
                    except redis.WatchError:
                        continue
        # Do not expire arguments of the first task since they have
        # already been deleted. However, all subsequent deferred task's
        # argument timeouts must be set.
        deferred_id = set_timeout(task_id, False)
        while deferred_id is not None:
            deferred_id = set_timeout(deferred_id)

    def get_status(self, task_id):
        key = TASK_PATTERN % (self.name, task_id)
        return self.redis.hget(key, 'status')

    def set_result(self, task_id, message, timeout):
        timeout = max(int(timeout), 1)
        task_key = TASK_PATTERN % (self.name, task_id)
        result_key = RESULT_PATTERN % (self.name, task_id)
        reserve_key = RESERVE_PATTERN % (self.name, task_id)
        with self.redis.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(reserve_key)
                    deferred_id = pipe.get(reserve_key)
                    pipe.multi()
                    if deferred_id is None:
                        pipe.hset(task_key, 'status', const.COMPLETED)
                        pipe.expire(task_key, timeout)
                        pipe.rpush(result_key, message)
                        pipe.expire(result_key, timeout)
                    else:
                        pipe.delete(task_key)
                        pipe.delete(result_key)
                    pipe.delete(reserve_key)
                    pipe.delete(ARGS_PATTERN % (self.name, task_id))
                    pipe.execute()
                    return deferred_id
                except redis.WatchError:
                    continue

    def pop_result(self, task_id, timeout):
        key = RESULT_PATTERN % (self.name, task_id)
        if timeout == 0:
            return self.redis.lpop(key)

        task = TASK_PATTERN % (self.name, task_id)
        end = None if timeout is None else time.time() + timeout
        while True:
            with self.redis.pipeline() as pipe:
                pipe.ttl(task)
                pipe.exists(task)
                timeout, exists = pipe.execute()
                if end:
                    timeout = min(timeout, int(end - time.time()))
            t = timeout
            if timeout <= 0:
                if not exists:
                    # task has expired (heartbeat stopped)
                    result = self.redis.lpop(key)
                    return const.TASK_EXPIRED if result is None else result
                # task is still enqueued (heartbeat not started yet)
                timeout = 1 if end else 5
            result = self.redis.blpop([key], timeout=timeout)
            if result is None:
                if end and time.time() > end:
                    return None # timeout
                continue
            self.redis.delete(task)
            return result[1]

    def discard_result(self, task_id, task_expired_token):
        result_key = RESULT_PATTERN % (self.name, task_id)
        task_key = TASK_PATTERN % (self.name, task_id)
        with self.redis.pipeline() as pipe:
            pipe.delete(task_key)
            pipe.rpush(result_key, task_expired_token)
            pipe.expire(result_key, 5)
            pipe.execute()



# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import worq.const as const
from worq import get_broker, get_queue, Task
from worq.task import Task, TaskFailure, TaskExpired
from worq.tests.util import (assert_raises, eq_, eventually, thread_worker,
    with_urls, TimeoutLock, WAIT)


@with_urls
def test_Broker_task_failed(url):
    lock = TimeoutLock(locked=True)
    def func():
        lock.acquire()
    broker = get_broker(url)
    broker.expose(func)
    with thread_worker(broker):
        q = get_queue(url)

        res = q.func()
        broker.task_failed(res)
        assert res.wait(timeout=WAIT), repr(res)
        lock.release()

        with assert_raises(TaskExpired):
            res.value


@with_urls
def test_Broker_duplicate_task_id_string(url):
    Broker_duplicate_task_id(url, 'int')

@with_urls
def test_Broker_duplicate_task_id_function(url):
    Broker_duplicate_task_id(url, lambda arg: type(arg).__name__)

def Broker_duplicate_task_id(url, identifier):
    lock = TimeoutLock(locked=True)
    state = []

    def func(arg):
        lock.acquire()
        return arg

    broker = get_broker(url)
    broker.expose(func)
    with thread_worker(broker, lock):
        q = get_queue(url)

        task = Task(q.func, id=identifier)
        res = task(1)

        eventually((lambda:res.status), const.ENQUEUED)
        msg = 'func [default:int] cannot enqueue task with duplicate id'
        with assert_raises(TaskFailure, msg):
            task(2)

        lock.release()
        eventually((lambda:res.status), const.PROCESSING)
        msg = 'func [default:int] cannot enqueue task with duplicate id'
        with assert_raises(TaskFailure, msg):
            task(3)

        lock.release()
        assert res.wait(timeout=WAIT), repr(res)
        eq_(res.value, 1)

        res = task(4)
        eventually((lambda:res.status), const.ENQUEUED)
        lock.release()
        eventually((lambda:res.status), const.PROCESSING)
        lock.release()
        assert res.wait(timeout=WAIT), repr(res)
        eq_(res.value, 4)



# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from worq import get_broker, get_queue, Task, TaskFailure, TaskSpace
from worq.tests.util import (assert_raises, eq_, eventually, thread_worker,
    with_urls, WAIT)


@with_urls
def test_worker_interrupted(url):

    def func(arg):
        raise KeyboardInterrupt()

    broker = get_broker(url)
    broker.expose(func)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        res = q.func('arg')
        completed = res.wait(WAIT)

        assert completed, repr(res)
        eq_(repr(res), '<Deferred func [default:%s] failed>' % res.id)
        with assert_raises(TaskFailure,
                'func [default:%s] KeyboardInterrupt: ' % res.id):
            res.value


@with_urls
def test_deferred_task_fail_on_error(url):

    def func(arg):
        if arg == 0:
            raise Exception('zero fail!')
        return arg

    broker = get_broker(url)
    broker.expose(func)
    with thread_worker(broker):

        # -- task-invoking code, usually another process --
        q = get_queue(url)

        res = q.func([q.func(1), q.func(0), q.func(2)])
        res.wait(timeout=WAIT)

        msg = 'func [default:%s] Exception: zero fail!' % res.task.args[0][1].id
        with assert_raises(TaskFailure, msg):
            res.value



# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import logging.config
import os
import re
import signal
import sys
from contextlib import contextmanager
from nose.tools import nottest
from os.path import dirname, exists, join
from worq import get_broker, get_queue
from worq.pool.process import WorkerPool, Error, run_in_subprocess
from worq.task import Task, TaskExpired
from worq.tests.util import (assert_raises, eq_, eventually, tempdir,
    with_urls, WAIT)

log = logging.getLogger(__name__)


@with_urls
def test_WorkerPool_start_twice(url):
    broker = get_broker(url)
    pool = WorkerPool(broker, get_broker, workers=1)
    with start_pool(pool):
        with assert_raises(Error):
            pool.start(handle_sigterm=False)


@with_urls
def test_WorkerPool_max_worker_tasks(url):
    broker = get_broker(url)
    pool = WorkerPool(broker, WorkerPool_max_worker_tasks_init,
        workers=1, max_worker_tasks=3)
    with start_pool(pool):

        q = get_queue(url)
        res = q.results([q.func() for n in range(4)])
        assert res.wait(WAIT), repr(res)

        results = res.value
        assert isinstance(results, list), results
        eq_([r[1] for r in results], [1, 2, 3, 1])
        eq_(len(set(r[0] for r in results)), 2)

def WorkerPool_max_worker_tasks_init(url):
    broker = get_broker(url)
    calls = [0]

    @broker.expose
    def func():
        calls[0] += 1
        return (os.getpid(), calls[0])

    @broker.expose
    def results(res):
        return res

    return broker


@with_urls
def test_WorkerPool_heartrate(url):
    broker = get_broker(url)
    pool = WorkerPool(broker, WorkerPool_heartrate_init, workers=1)
    with start_pool(pool):

        q = get_queue(url)

        res = Task(q.suicide_worker, heartrate=0.1, result_timeout=5)()
        assert res.wait(WAIT), repr(res)
        print repr(res)
        with assert_raises(TaskExpired):
            res.value

def WorkerPool_heartrate_init(url):
    broker = get_broker(url)

    @broker.expose
    def suicide_worker():
        log.warn("it's nice to work alone")
        os.kill(os.getpid(), signal.SIGKILL) # force kill worker

    broker.expose(os.getpid)

    return broker


@nottest # this is a very slow test, and doesn't seem that important
@with_urls(exclude='memory')
def test_WorkerPool_crashed_worker(url):
    broker = get_broker(url)
    pool = WorkerPool(broker, WorkerPool_crashed_worker_init, workers=1)
    with start_pool(pool):

        q = get_queue(url)

        res = q.getpid()
        assert res.wait(WAIT), repr(res)
        pid = res.value

        q.kill_worker()

        res = q.getpid()
        assert res.wait(WAIT), repr(res)
        assert res.value != pid, pid

def WorkerPool_crashed_worker_init(url):
    broker = get_broker(url)

    @broker.expose
    def kill_worker():
        log.warn('alone we crash')
        sys.exit()

    broker.expose(os.getpid)

    return broker


@with_urls(exclude='memory')
def test_WorkerPool_sigterm(url):
    with tempdir() as tmp:

        logpath = join(tmp, 'output.log')
        proc = run_in_subprocess(worker_pool, url,
            WorkerPool_sigterm_init, (tmp, logpath), workers=3)

        with printlog(logpath), force_kill_on_exit(proc):

            q = get_queue(url)

            q.func('text')

            eventually(reader(tmp, 'func.started'), '')

            proc.terminate() # signal pool shutdown
            touch(join(tmp, 'func.unlock')) # allow func to proceed

            eventually(reader(tmp, 'func.out'), 'text')
            eventually(verify_shutdown(proc), True, timeout=WAIT)

def WorkerPool_sigterm_init(url, tmp, logpath):
    process_config(logpath, 'Worker-%s' % os.getpid())
    broker = get_broker(url)

    @broker.expose
    def func(arg, lock=None):
        # signal func started
        touch(join(tmp, 'func.started'))

        # wait for unlock
        eventually(reader(tmp, 'func.unlock'), '')

        # write data to file
        touch(join(tmp, 'func.out'), arg)

        log.debug('func complete')

    return broker


@with_urls(exclude='memory')
def test_WorkerPool_worker_shutdown_on_parent_die(url):
    with tempdir() as tmp:

        logpath = join(tmp, 'output.log')
        proc = run_in_subprocess(worker_pool, url,
            WorkerPool_worker_shutdown_on_parent_die_init,
            (tmp, logpath))

        with printlog(logpath), force_kill_on_exit(proc):

            res = get_queue(url).getpid()
            assert res.wait(WAIT), repr(res)

            os.kill(proc.pid, signal.SIGKILL) # force kill pool master
            eventually(proc.is_alive, False, timeout=WAIT)

        try:
            eventually(pid_running(res.value), False,
                timeout=WAIT, poll_interval=0.1)
        except Exception:
            os.kill(res.value, signal.SIGTERM) # clean up
            raise

def WorkerPool_worker_shutdown_on_parent_die_init(url, tmp, logpath):
    process_config(logpath, 'Worker-%s' % os.getpid())
    broker = get_broker(url)

    broker.expose(os.getpid)

    import worq.pool.process as proc
    proc.WORKER_POLL_INTERVAL = 0.1

    return broker

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# pool test helpers

def process_config(logpath, procname):
    import worq.pool.process as proc
    proc.WORKER_POLL_INTERVAL = 1
    logging.config.dictConfig({
        'formatters': {
            'brief': {
                'format': ('%(asctime)s ' + procname
                    + ' %(levelname)-7s %(name)s - %(message)s'),
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.FileHandler',
                'filename': logpath,
                'mode': 'a',
                'formatter': 'brief',
            },
        },
        'root': {
            'level': 'DEBUG',
            'handlers': ['console'],
        },
        'disable_existing_loggers': False,
        'version': 1,
    })
    log.info('WORKER_POLL_INTERVAL = %s', proc.WORKER_POLL_INTERVAL)

def touch(path, data=''):
    with open(path, 'w') as f:
        f.write(data)

def reader(*path):
    path = join(*path)
    def read():
        if exists(path):
            with open(path) as f:
                return f.read()
    return read

def verify_shutdown(proc):
    def verify():
        return not proc.is_alive()
    return verify

def pid_running(pid):
    def is_process_running(pid=pid):
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        return True
    return is_process_running

@contextmanager
def printlog(logpath, heading='pool logging output:'):
    try:
        yield
    finally:
        if exists(logpath):
            print heading
            with open(logpath) as f:
                print f.read(),

def _logging_init(url, _logpath, _init, *args, **kw):
    if exists(dirname(_logpath)):
        process_config(_logpath, 'Worker-%s' % os.getpid())
    else:
        # worker was probably orphaned
        sys.exit()
    return _init(url, *args, **kw)

def worker_pool(url, init_func, init_args, workers=1):
    process_config(init_args[-1], 'Pool-%s' % os.getpid())
    broker = get_broker(url)
    with discard_tasks(broker):
        pool = WorkerPool(broker, init_func, init_args, workers=workers)
        pool.start(timeout=1)

@contextmanager
def start_pool(pool, timeout=1):
    with tempdir() as tmp:
        assert exists(tmp), tmp
        logpath = join(tmp, 'log')
        pool.init_kwargs.setdefault('_logpath', logpath)
        pool.init_kwargs.setdefault('_init', pool.init_func)
        pool.init_func = _logging_init
        with discard_tasks(pool.broker), printlog(logpath):
            pool.start(timeout=timeout, handle_sigterm=False)
            try:
                yield
            finally:
                pool.stop()

@contextmanager
def discard_tasks(broker):
    try:
        yield
    finally:
        broker.discard_pending_tasks()

@contextmanager
def force_kill_on_exit(proc):
    try:
        yield
    except:
        log.error('original error', exc_info=True)
        raise
    finally:
        if proc.is_alive():
            proc.join(WAIT)
            if proc.is_alive():
                # force kill
                os.kill(proc.pid, signal.SIGKILL)
                raise RuntimeError('Had to force kill broker process. '
                    'Worker subprocesses may be orphaned.')



# WorQ - Python task queue
#
# Copyright (c) 2012 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import
from urlparse import urlparse
from worq.core import DEFAULT, Broker
from worq.task import Task, TaskFailure, TaskSpace
from worq.queue.memory import TaskQueue as MemoryQueue
try:
    from worq.queue.redis import TaskQueue as RedisQueue
except ImportError:
    RedisQueue = None

__version__ = '1.0.0'

BROKER_REGISTRY = {'memory': MemoryQueue.factory}

if RedisQueue is not None:
    BROKER_REGISTRY['redis'] = RedisQueue

def get_broker(url, name=DEFAULT, *args, **kw):
    """Create a new broker

    :param url: Task queue URL.
    :param name: The name of the queue on which to expose or invoke tasks.
    :returns: An instance of ``worq.core.Broker``.
    """
    scheme = urlparse(url).scheme
    try:
        factory = BROKER_REGISTRY[scheme]
    except KeyError:
        raise ValueError('invalid broker URL: %s' % url)
    return Broker(factory(url, name, *args, **kw))

def get_queue(url, name=DEFAULT, target=''):
    """Get a queue for invoking remote tasks

    :param url: Task queue URL.
    :param name: The name of the queue on which tasks should be invoked.
        Queued tasks will be invoked iff there is a worker listening on the
        named queue.
    :param target: Task namespace (similar to a python module) or name
        (similar to a python function). Defaults to the root namespace.
    :returns: An instance of ``worq.task.Queue``.
    """
    return get_broker(url, name).queue(target)



from setuptools import setup, find_packages
from worq import __version__

setup(
    name='WorQ',
    version=__version__,
    author='Daniel Miller',
    author_email='millerdev@gmail.com',
    packages=find_packages(),
    url='http://worq.readthedocs.org/',
    license='LICENSE.txt',
    description='Python task queue',
    long_description=open('README.rst').read(),
    extras_require={
        "redis": "redis >= 2.4",
    },
)
